/*
 * Driver for scn2681 duart; Cisco 2500 console and clock tick.
 * Koen De Vleeschauwer, 2000.
 * Derived from mon960_console.c,v 1.1.2.9 1999/05/10 20:26:33 kma
 *
 */

#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/config.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/mm.h>
#include <linux/kernel.h>
 
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <asm/bitops.h>
#include <asm/delay.h>     

#include <asm/cisco2500.h>
#include "2681serial.h"

static int scn_refcount;
static struct tty_struct *scn_table[1];
static struct termios *scn_termios[1];
static struct termios *scn_termios_locked[1];
struct tty_driver scn_driver;

void rs_put_char(char);

static int scn_write(struct tty_struct* tty, int from_user,
		const unsigned char* buf, int count)
{
	int charstowrite;
	/* real = virtual in uClinux, hence no looking at from_user */
	charstowrite = count;	
	while (charstowrite--)
		rs_put_char(*buf++);
	return count;
}

static int scn_chars_in_buffer(struct tty_struct *tty)
{
        return 0;               /* we're not buffering */
}            

static int scn_writeroom(struct tty_struct *tty)
{
        return  8192;           /* could return anything */
}

static void scn_putchar(struct tty_struct* tty, unsigned char c)
{
        rs_put_char(c);
}

static int scn_open(struct tty_struct* tty, struct file* filp)
{
        return 0;
}                                                                               

void scn_init(void)
{
	printk("2681 serial console");

	memset(&scn_driver, 0, sizeof(scn_driver));
	scn_driver.magic = TTY_DRIVER_MAGIC;
	scn_driver.name = "ttyS";
	scn_driver.name_base = 1;
	scn_driver.major = TTY_MAJOR;
	scn_driver.minor_start = 64;
	scn_driver.num = 1;
	scn_driver.type = TTY_DRIVER_TYPE_SERIAL;
	scn_driver.init_termios = tty_std_termios;
	scn_driver.init_termios.c_lflag |= ISIG | ICANON | ECHO 
		 | ECHOE | ECHOK ;
	scn_driver.init_termios.c_iflag |= ICRNL;
	scn_driver.init_termios.c_cflag = CS8 | CREAD | CLOCAL;
	scn_driver.flags = TTY_DRIVER_REAL_RAW;
	scn_driver.refcount = &scn_refcount;

	scn_driver.table = scn_table;
	scn_driver.termios = scn_termios;
	scn_driver.termios_locked = scn_termios_locked;

	scn_driver.open = scn_open;
	scn_driver.write = scn_write;
	scn_driver.put_char = scn_putchar;
	scn_driver.write_room = scn_writeroom;
	scn_driver.chars_in_buffer = scn_chars_in_buffer;

	if (tty_register_driver(&scn_driver))
		printk("Couldn't register console driver\n");

	printk(" initialized\n");	
}

/****************************************************************************/
/*                          100Hz Clock tick                                */
/****************************************************************************/  

void scn_tick(void)
{
        volatile unsigned char  *uartp;
        volatile unsigned char  dummy;
	struct tty_struct	*tty = scn_table[0];
        unsigned char buf[256];
	int charsinbuf;

	/*
	 * XXX fixme. As the clock interrupt is also 
	 * used for console input, clock is off by 1/100 sec
	 * per character received :(
	 */
 
        uartp = (volatile unsigned char *) DUARTBASE ;

	/* timer interrupt ? */
       	if (uartp[CHAN_A + CH_SR] & INT_CTR)
        	dummy = uartp[DU_CSTOP];

	/* receive chars ? */
	for (charsinbuf=0; uartp[CHAN_A + CH_SR] & SR_RX_RDY; charsinbuf++ ) {
		buf[charsinbuf] = uartp[CHAN_A + CH_DAT] ;
		}
	if (charsinbuf)
		tty->ldisc.receive_buf(tty, buf, 0, charsinbuf);
}            

void scn_timer_init(void (*handler)(int, void *, struct pt_regs *))
{
        volatile unsigned char *uartp;
        volatile unsigned char  dummy;
        unsigned short timercycle;
 
        /* Set up timer clock tick */
        uartp = (volatile unsigned char *) DUARTBASE;
 
        uartp[DU_IMR] = INT_NONE ;
        dummy = uartp[DU_CSTOP] ;
 
        /* set up timer counter for 100 Hz tick */
        uartp[DU_ACR] = ACR_CT_TCLK16 ; /* timer mode, clock : 16 */
        timercycle = (unsigned short) ((DU_CLK / 32) / HZ);
        uartp[DU_CTUR] = timercycle >> 8;
        uartp[DU_CTLR] = timercycle && 0xff ;
 
        /* register irq handler. 29 is interrupt 5, no autovector,
           as 2681 can't autovector (68681 can) */
        request_irq(29, handler, SA_INTERRUPT, "Console duart", NULL);
        /* enable clock timer interrupts */
        uartp[DU_IMR] = INT_CTR + INT_RXA ;
        dummy = uartp[DU_CSTRT];
}                   

/* disable uart interrupts, used by trap_init */
void scn_nointerrupts(void)
{
	volatile unsigned char  *uartp;
	uartp = (volatile unsigned char *) DUARTBASE ;

        uartp[DU_IMR] = INT_NONE ;
}


/****************************************************************************/
/*                          Serial Console                                  */
/****************************************************************************/  
/*
 *	Output a single character, using UART polled mode.
 *	This is used for console output.
 *	Note: printk blocks the kernel until everything has been printed 
 *	Use sparingly.
 */

void rs_put_char(char ch)
{
	volatile unsigned char  *uartp;
	uartp = (volatile unsigned char *) DUARTBASE ;
	
	while (!(uartp[CHAN_A+CH_SR] & SR_TX_RDY))
		;
	uartp[CHAN_A+CH_DAT] = (unsigned char)ch;
	while (!(uartp[CHAN_A+CH_SR] & SR_TX_RDY))
		;
	return;
}

/*
 * rs_console_print is registered for printk output, early in the boot process.
 */

void rs_console_print(const char *p)
{
	char c;
	
	while ((c = *(p++)) != 0) {
		if(c == '\n')
			rs_put_char('\r');
		rs_put_char(c);
	}
	return;
}

/****************************************************************************/
