/*****************************************************************************/

/*
 *	play -- play "wav" files, through the dac0800 driver.
 *
 *	(C) Copyright 1999, Greg Ungerer (gerg@moreton.com.au)
 */

/*****************************************************************************/

#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <getopt.h>

#include "wav.h"

/*****************************************************************************/

#define	DACDEVICE	"/dev/dac"

int			verbose = 1;
int			raw = 0;
unsigned char		buf[16*1024];

struct wavefileheader	hdr;
struct wavedataheader	datahdr;


/*
 *	Endian re-arranger :-)
 */
#define	SWAPLONG(a)	((((a) & 0xff000000) >> 24) | \
			(((a) & 0x00ff0000) >> 8) | \
			(((a) & 0x0000ff00) << 8) | \
			(((a) & 0x000000ff) << 24))

#define	SWAPSHORT(a)	(((a) >> 8) | ((a) << 8))

/*****************************************************************************/

void usage(void)
{
	printf("usage: play [-vrsw] [-f frequency] filename\n");
	exit(1);
}

/*****************************************************************************/

int main(int argc, char *argv[])
{
	char	*filename;
	int	ofd, ifd, i, j, n;
	int	freq, chans, bitspersample;
	char	c;

	freq = 0;
	chans = 1;
	bitspersample = 8;

	while ((c = getopt(argc, argv, "vrswf:")) > 0) {
		switch (c) {
		case 'v':
			verbose++;
			break;
		case 'f':
			freq = atoi(optarg);
			break;
		case 'r':
			if (freq == 0)
				freq = 8000;
			raw++;
			break;
		case 's':
			chans = 2;
			break;
		case 'w':
			bitspersample = 16;
			break;
		default:
			usage();
			break;
		}
	}

	if (optind >= argc)
		usage();

	filename = argv[optind];
	if ((ifd = open(filename, O_RDONLY)) < 1) {
		printf("ERROR: failed to wave file %s\n", filename);
		exit(1);
	}
	if ((ofd = open(DACDEVICE, O_RDWR)) < 1) {
		printf("ERROR: failed to open DAC device %s\n", DACDEVICE);
		exit(1);
	}

	if (!raw) {
		/* Parse wav header... */
		if ((n = read(ifd, &hdr, sizeof(hdr))) != sizeof(hdr)) {
			printf("ERROR: failed to read(), errno=%d\n", errno);
			exit(1);
		}

		hdr.filesize = SWAPLONG(hdr.filesize);
		hdr.formattag = SWAPSHORT(hdr.formattag);
		hdr.channels = SWAPSHORT(hdr.channels);
		hdr.samplerate = SWAPLONG(hdr.samplerate);
		hdr.bytespersec = SWAPLONG(hdr.bytespersec);
		hdr.samplesize = SWAPSHORT(hdr.samplesize);
		hdr.bitspersample = SWAPSHORT(hdr.bitspersample);

		if (verbose) {
			printf("FILE=%s\n  Type=0x%x\n  File Length=%d\n",
				filename, hdr.riffid, hdr.filesize);
			printf("  Channels=%d\n  Sample Rate=%d\n  Bytes Per "
				"Second=%d\n  Sample Size=%d\n  "
				"Bits per Sample=%d\n", hdr.channels,
				hdr.samplerate, hdr.bytespersec,
				hdr.samplesize, hdr.bitspersample);
		}
	
		if (hdr.riffid != WAV_RIFFTYPE) {
			printf("ERROR: unknown file type, riffid=%x??\n",
				hdr.riffid);
			exit(1);
		}
		if (hdr.waveid != WAV_WAVETYPE) {
			printf("ERROR: unknown file type, waveid=%x??\n",
				hdr.waveid);
			exit(1);
		}
		if (hdr.fmtid != WAV_FMTTYPE) {
			printf("ERROR: unknown file type, fmtid=%x??\n",
				hdr.fmtid);
			exit(1);
		}

		chans = hdr.channels;
		bitspersample = hdr.bitspersample;
	}

	/* Set the sample rate */
	if (freq == 0)
		freq = hdr.samplerate;
	if (ioctl(ofd, 1, freq) < 0) {
		printf("ERROR: ioctl(freq) failed, errno=%d\n",
			errno);
		exit(1);
	}

	/* Parse data header, and ignore... */
	if ((n = read(ifd, &datahdr, sizeof(datahdr))) != sizeof(datahdr)) {
		printf("ERROR: failed to read(), errno=%d\n", errno);
		exit(1);
	}

	/* Output data */
	while ((n = read(ifd, buf, sizeof(buf))) > 0) {
		if (bitspersample == 16) {
			/* Make audio track 8 bits only */
			/*for (i = 2, j = 1; (i < n); i += 2, j++)*/
			for (i = 1, j = 0; (i < n); i += 2, j++)
				buf[j] = buf[i];
			n = j;
		}
		if (chans == 2) {
			/* Make audio track single channel only */
			for (i = 2, j = 1; (i < n); i += 2, j++)
				buf[j] = buf[i];
			n = j;
		}

		if (write(ofd, buf, n) != n) {
			printf("ERROR: write(%s) failed, errno=%d\n",
				DACDEVICE, errno);
		}
	}

	close(ofd);
	close(ifd);
	exit(0);
}

/*****************************************************************************/
